/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
//gl_fog.c -- global and volumetric fog

#include "quakedef.h"

//==============================================================================
//
//  GLOBAL FOG
//
//==============================================================================

float fog_density;
float fog_start; // Kurok
float fog_end;   // Kurok
float fog_red;
float fog_green;
float fog_blue;

float old_density;
float old_start; // Kurok
float old_end;   // Kurok
float old_red;
float old_green;
float old_blue;

float fade_time; //duration of fade
float fade_done; //time when fade will be done


/*
=============
Fog_Update

update internal variables
=============
*/
static void Fog_Update_Kurok (float start, float end, float red, float green, float blue, float time)
{
	//save previous settings for fade
	if (time > 0)
	{
		//check for a fade in progress
		if (fade_done > cl.time)
		{
			float f, d;

			f = (fade_done - cl.time) / fade_time;
			old_start = f * old_start + (1.0 - f) * fog_start;
			old_end = f * old_end + (1.0 - f) * fog_end;
			old_red = f * old_red + (1.0 - f) * fog_red;
			old_green = f * old_green + (1.0 - f) * fog_green;
			old_blue = f * old_blue + (1.0 - f) * fog_blue;
		}
		else
		{
			old_start = fog_start;
			old_end = fog_end;
			old_red = fog_red;
			old_green = fog_green;
			old_blue = fog_blue;
		}
	}

	fog_start = start;
	fog_end = end;
	fog_red = red;
	fog_green = green;
	fog_blue = blue;
	fade_time = time;
	fade_done = cl.time + time;
}

/*
=============
Fog_Update

update internal variables
=============
*/
static void Fog_Update (float density, float red, float green, float blue, float time)
{
	//save previous settings for fade
	if (time > 0)
	{
		//check for a fade in progress
		if (fade_done > cl.time)
		{
			float f;//, d; unused -- kristian

			f = (fade_done - cl.time) / fade_time;
			old_density = f * old_density + (1.0 - f) * fog_density;
			old_red = f * old_red + (1.0 - f) * fog_red;
			old_green = f * old_green + (1.0 - f) * fog_green;
			old_blue = f * old_blue + (1.0 - f) * fog_blue;
		}
		else
		{
			old_density = fog_density;
			old_red = fog_red;
			old_green = fog_green;
			old_blue = fog_blue;
		}
	}

	fog_density = density;
	fog_red = red;
	fog_green = green;
	fog_blue = blue;
	fade_time = time;
	fade_done = cl.time + time;
}

/*
=============
Fog_ParseServerMessage

handle an SVC_FOG message from server
=============
*/
void Fog_ParseServerMessage (void)
{
	float start, end; // Kurok
	float density, red, green, blue, time;

	if (kurok)
    {
		start = MSG_ReadByte() / 255.0;
		end = MSG_ReadByte() / 255.0;
    }
	else
		density = MSG_ReadByte() / 255.0;
	red = MSG_ReadByte() / 255.0;
	green = MSG_ReadByte() / 255.0;
	blue = MSG_ReadByte() / 255.0;
	time = max(0.0, MSG_ReadShort() / 100.0);

	if (kurok)
		Fog_Update_Kurok (start, end, red, green, blue, time);
	else
		Fog_Update (density, red, green, blue, time);
}

/*
=============
Fog_FogCommand_f

handle the 'fog' console command
=============
*/
static void Fog_FogCommand_f (void)
{
	if (kurok)
	{
		switch (Cmd_Argc())
		{
		default:
		case 1:
			Con_Printf("usage:\n");
			Con_Printf("   fog <fade>\n");
			Con_Printf("   fog <start> <end>\n");
			Con_Printf("   fog <red> <green> <blue>\n");
			Con_Printf("   fog <fade> <red> <green> <blue>\n");
			Con_Printf("   fog <start> <end> <red> <green> <blue>\n");
			Con_Printf("   fog <start> <end> <red> <green> <blue> <fade>\n");
			Con_Printf("current values:\n");
			Con_Printf("   \"start\" is \"%f\"\n", fog_start);
			Con_Printf("   \"end\" is \"%f\"\n", fog_end);
			Con_Printf("   \"red\" is \"%f\"\n", fog_red);
			Con_Printf("   \"green\" is \"%f\"\n", fog_green);
			Con_Printf("   \"blue\" is \"%f\"\n", fog_blue);
			Con_Printf("   \"fade\" is \"%f\"\n", fade_time);
			break;
		case 2: //TEST
			Fog_Update_Kurok(fog_start,
					   fog_end,
					   fog_red,
					   fog_green,
					   fog_blue,
					   0.0);
			break;
		case 3:
			Fog_Update_Kurok(atof(Cmd_Argv(1)),
					   atof(Cmd_Argv(2)),
					   fog_red,
					   fog_green,
					   fog_blue,
					   0.0);
			break;
		case 4:
			Fog_Update_Kurok(fog_start,
					   fog_end,
					   CLAMP(0.0, atof(Cmd_Argv(1)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(2)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(3)), 100.0),
					   0.0);
			break;
		case 5: //TEST
			Fog_Update_Kurok(fog_start,
					   fog_end,
					   CLAMP(0.0, atof(Cmd_Argv(1)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(2)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(3)), 100.0),
					   0.0);
			break;
		case 6:
			Fog_Update_Kurok(atof(Cmd_Argv(1)),
					   atof(Cmd_Argv(2)),
					   CLAMP(0.0, atof(Cmd_Argv(3)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(4)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(5)), 100.0),
					   0.0);
			break;
		case 7:
			Fog_Update_Kurok(atof(Cmd_Argv(1)),
					   atof(Cmd_Argv(2)),
					   CLAMP(0.0, atof(Cmd_Argv(3)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(4)), 100.0),
					   CLAMP(0.0, atof(Cmd_Argv(5)), 100.0),
					   0.0);
			break;
		}

		return;
	}

	// Normal use starts here
	switch (Cmd_Argc())
	{
	default:
	case 1:
		Con_Printf("usage:\n");
		Con_Printf("   fog <density>\n");
		Con_Printf("   fog <red> <green> <blue>\n");
		Con_Printf("   fog <density> <red> <green> <blue>\n");
		Con_Printf("current values:\n");
		Con_Printf("   \"density\" is \"%f\"\n", fog_density);
		Con_Printf("   \"red\" is \"%f\"\n", fog_red);
		Con_Printf("   \"green\" is \"%f\"\n", fog_green);
		Con_Printf("   \"blue\" is \"%f\"\n", fog_blue);
		break;
	case 2:
		Fog_Update(max(0.0, atof(Cmd_Argv(1))),
				   fog_red,
				   fog_green,
				   fog_blue,
				   0.0);
		break;
	case 3: //TEST
		Fog_Update(max(0.0, atof(Cmd_Argv(1))),
				   fog_red,
				   fog_green,
				   fog_blue,
				   atof(Cmd_Argv(2)));
		break;
	case 4:
		Fog_Update(fog_density,
				   CLAMP(0.0, atof(Cmd_Argv(1)), 1.0),
				   CLAMP(0.0, atof(Cmd_Argv(2)), 1.0),
				   CLAMP(0.0, atof(Cmd_Argv(3)), 1.0),
				   0.0);
		break;
	case 5:
		Fog_Update(fmax(0.0, atof(Cmd_Argv(1))),
				   CLAMP(0.0, atof(Cmd_Argv(2)), 1.0),
				   CLAMP(0.0, atof(Cmd_Argv(3)), 1.0),
				   CLAMP(0.0, atof(Cmd_Argv(4)), 1.0),
				   0.0);
		break;
	case 6: //TEST
		Fog_Update(fmax(0.0, atof(Cmd_Argv(1))),
				   CLAMP(0.0, atof(Cmd_Argv(2)), 1.0),
				   CLAMP(0.0, atof(Cmd_Argv(3)), 1.0),
				   CLAMP(0.0, atof(Cmd_Argv(4)), 1.0),
				   atof(Cmd_Argv(5)));
		break;
	}
}

/*
=============
Fog_ParseWorldspawn

called at map load
=============
*/
static void Fog_ParseWorldspawn (void)
{
	char key[128], value[4096];
	char *data;

	//initially no fog
	fog_density = 0.0;
	old_density = 0.0;
	// Kurok starts here
	fog_start = 0.0;
	old_start = 0.0;

	fog_end = 0.0;
	old_end = 0.0;

	fog_red = 0.0;
	old_red = 0.0;

	fog_green = 0.0;
	old_green = 0.0;

	fog_blue = 0.0;
	old_blue = 0.0;
	// End Kurok
	fade_time = 0.0;
	fade_done = 0.0;

	data = COM_Parse(cl.worldmodel->entities);
	if (!data)
		return; // error
	if (com_token[0] != '{')
		return; // error
	while (1)
	{
		data = COM_Parse(data);
		if (!data)
			return; // error
		if (com_token[0] == '}')
			break; // end of worldspawn
		if (com_token[0] == '_')
			strcpy(key, com_token + 1);
		else
			strcpy(key, com_token);
		while (key[strlen(key)-1] == ' ') // remove trailing spaces
			key[strlen(key)-1] = 0;
		data = COM_Parse(data);
		if (!data)
			return; // error
		strcpy(value, com_token);

		if (!strcmp("fog", key))
		{
			if (kurok)
				sscanf(value, "%f %f %f %f %f", &fog_start, &fog_end, &fog_red, &fog_green, &fog_blue);

			else
				sscanf(value, "%f %f %f %f", &fog_density, &fog_red, &fog_green, &fog_blue);
		}
	}
}

/*
=============
Fog_GetColor

calculates fog color for this frame, taking into account fade times
=============
*/
float s, e;
float *Fog_GetColor (void)
{
	static float c[4];
	float f;
	int i;

	if (kurok)
	{
		if (fade_done > cl.time)
		{
			f = (fade_done - cl.time) / fade_time;
			s = f * old_start + (1.0 - f) * fog_start;
			e = f * old_end + (1.0 - f) * fog_end;
			c[0] = f * old_red + (1.0 - f) * fog_red * 0.01;
			c[1] = f * old_green + (1.0 - f) * fog_green * 0.01;
			c[2] = f * old_blue + (1.0 - f) * fog_blue * 0.01;
			c[3] = 1.0;
		}
		else
		{
			s = fog_start;
			e = fog_end;
			c[0] = fog_red * 0.01;
			c[1] = fog_green * 0.01;
			c[2] = fog_blue * 0.01;
			c[3] = 1.0;
		}




	}
	else
	{


		if (fade_done > cl.time)
		{
			f = (fade_done - cl.time) / fade_time;
			c[0] = f * old_red + (1.0 - f) * fog_red;
			c[1] = f * old_green + (1.0 - f) * fog_green;
			c[2] = f * old_blue + (1.0 - f) * fog_blue;
			c[3] = 1.0;
		}
		else
		{
			c[0] = fog_red;
			c[1] = fog_green;
			c[2] = fog_blue;
			c[3] = 1.0;
		}
	}

	//find closest 24-bit RGB value, so solid-colored sky can match the fog perfectly
	for (i=0;i<3;i++)
		c[i] = (float)(Q_rint(c[i] * 255)) / 255.0f;


	return c;
}

/*
=============
Fog_GetDensity

returns current density of fog
=============
*/
float Fog_GetDensity (void)
{
	float f;

	if (kurok)
	{
		if (fade_done > cl.time)
		{
			f = (fade_done - cl.time) / fade_time;
			return f * old_end + (1.0 - f) * fog_end;
		}
		else
			return fog_end;

		return;
	}


	if (fade_done > cl.time)
	{
		f = (fade_done - cl.time) / fade_time;
		return f * old_density + (1.0 - f) * fog_density;
	}
	else
		return fog_density;
}

/*
=============
Fog_SetupFrame

called at the beginning of each frame
=============
*/
void Fog_SetupFrame (void)
{

	if (kurok)
	{
		glFogfv(GL_FOG_COLOR, Fog_GetColor());
        glFogf(GL_FOG_START, s);
		glFogf(GL_FOG_END, e);
	}
	else
	{
		glFogfv(GL_FOG_COLOR, Fog_GetColor());
		glFogf(GL_FOG_DENSITY, Fog_GetDensity() / 64.0);
	}
}

/*
=============
Fog_EnableGFog

called before drawing stuff that should be fogged
=============
*/
void Fog_EnableGFog (void)
{
	if (Fog_GetDensity() > 0)
		glEnable(GL_FOG);
}

/*
=============
Fog_DisableGFog

called after drawing stuff that should be fogged
=============
*/
void Fog_DisableGFog (void)
{
	if (Fog_GetDensity() > 0)
		glDisable(GL_FOG);
}

/*
=============
Fog_StartAdditive

called before drawing stuff that is additive blended -- sets fog color to black
=============
*/
void Fog_StartAdditive (void)
{
	vec3_t color = {0,0,0};

	if (Fog_GetDensity() > 0)
		glFogfv(GL_FOG_COLOR, color);
	}

/*
=============
Fog_StopAdditive

called after drawing stuff that is additive blended -- restores fog color
=============
*/
void Fog_StopAdditive (void)
	{
	if (Fog_GetDensity() > 0)
		glFogfv(GL_FOG_COLOR, Fog_GetColor());
}

//==============================================================================
//
//  VOLUMETRIC FOG
//
//==============================================================================

cvar_t r_vfog = {"r_vfog", "1"};

void Fog_DrawVFog (void){}
void Fog_MarkModels (void){}

//==============================================================================
//
//  INIT
//
//==============================================================================

/*
=============
Fog_NewMap

called whenever a map is loaded
=============
*/
void Fog_NewMap (void)
{
	Fog_ParseWorldspawn (); //for global fog
	Fog_MarkModels (); //for volumetric fog
}

void Fog_Reset (void)
{
	//set up global fog
	fog_density = 0.0;
	fog_start = 0.0; // Kurok
	fog_end = 0.0;   // kurok
	fog_red = 0.3;
	fog_green = 0.3;
	fog_blue = 0.3;

	if (kurok)
	{
		glFogi(GL_FOG_START, 0);
		glFogi(GL_FOG_MODE, GL_LINEAR);
	}
	else
		glFogi(GL_FOG_MODE, GL_EXP2);
	return;
}

/*
=============
Fog_Init

called when quake initializes
=============
*/
void Fog_Init (void)
{
	Cmd_AddCommand ("fog",Fog_FogCommand_f);
	Fog_Reset ();

}
